function MG_Zebra_y(n)

% Solve the 2D anisotropic diffusion problem,
% L(u) = - epsi*u_xx - u_yy = f, with Dirichlet boundary conditions,
% on a square grid with V(nu1,nu2) cycle using Zebra relaxation in the y direction. 
% Plot residual norm histories and convergence factor per cycle.
% The input n is the number of grid intervals in each direction. It
% should be a power of 2.

% Important Note: The solution u is of size (n+1) by (n+1). For convenience, 
% the right-hand side and L(u,h) are also set to this size, though only
% their interior values are defined. Their boundary values *must* be 0.  

n = 2^floor(log2(n))              % Reduce n to a power of 2
h = 1/n;                          % Mesh-size (the domain is the unit square)
nu1 = 1;                          % # Pre-relaxation sweeps
nu2 = 1;                          % # Post-relaxation sweeps
epsi = 100;                       % Anisotropy parameter
n_cycles = 10;                    % Number of cycles
[u,f] = Initialize_Data(n,h,epsi);% Define right-hand side, bounday conditions,
                                  % and initial guess for the solution
r = f - L(u,h,epsi);                   % The initial residual
rnorm = norm(r(:))/n;             % This will store the 
                                  % residual vector norm after each cycle 

                                  % Run V-Cycle iterations
for iters = 1:n_cycles,
    u = V_cycle(u,f,h,nu1,nu2,epsi);
    r = f - L(u,h,epsi);
    rnorm = [rnorm, norm(r(:))/n];
end % for

% Plot residual norm histories on a Log(y) scale
figure(1)
semilogy(0:n_cycles,rnorm,'ko-');
title('Residual Norm History')
xlabel('Iterations')
ylabel('Residual Norm')

% Plot residual norm convergence factors
figure(2)
plot(1:n_cycles,rnorm(2:end) ./ rnorm(1:end-1),'ko-');
title('Residual Convergence Factor Per Cycle')
xlabel('Iterations')
ylabel('Convergence Factor')

figure(3)
surf(u)
title('Final Solution u')

return


function Lu = L(u,h,epsi) % Computes L(u)

Lu = zeros(size(u)); % Lu is the same size as u for convenience. Only the
                     % interior values are defined.
Lu(2:end-1,2:end-1) = -(u(1:end-2,2:end-1) + u(3:end,2:end-1) + ...
    epsi*u(2:end-1,1:end-2) + epsi*u(2:end-1,3:end) - 2*(1+epsi)*u(2:end-1,2:end-1))/h^2;
return;


function u = V_cycle(u,f,h,nu1,nu2,epsi)   % Performs a V cycle; recursive.

if (mod(length(u),2) == 0)              % Verify even number of intervals.
    disp('Make n a power of 2');
    return;
end % if

if (length(u) <= 3)                     % Coarsest grid
    u = Jacobi(u,f,h,1,1,epsi);         % One undamped Jacobi solves exactly, 
                                        % because there is only one variable.
    return;
end % if

u = Zebra_y(u,f,h,nu1,epsi);            % Perform nu1 Zebra_y relaxation sweeps

% Compute residual
r = f - L(u,h,epsi);

fc = Restrict(r);                       % Right-hand side of coarser grid
uc = zeros(size(fc));                   % Zero initial guess on coarse grid
uc = V_cycle(uc,fc,2*h,nu1,nu2,epsi);        % Recursive call
u = u + Prolong(uc);                    % Interpolate and add correction

u = Zebra_y(u,f,h,nu1,epsi);            % Perform nu2 Zebra_y relaxation sweeps

return

function u = Zebra_y(u,f,h,nu,epsi)               % Zebra_y relaxation 

m = size(u,2); % total number of columns, including boundary
Diag_Term = 2*epsi/h^2;
Red_Stripes = 2:2:m-1;
Black_Stripes = 3:2:m-2;
for iterations = 1:nu
    rhs = f(:,Red_Stripes);
    rhs(2:end-1,:) = f(2:end-1,Red_Stripes) + ...
        epsi*(u(2:end-1,Red_Stripes+1) + u(2:end-1,Red_Stripes-1))/h^2;
    u(:,Red_Stripes) = Relax_Stripes_y(u(:,Red_Stripes),rhs,h,Diag_Term);
    rhs = f(:,Black_Stripes);
    rhs(2:end-1,:) = f(2:end-1,Black_Stripes) + ...
        epsi*(u(2:end-1,Black_Stripes+1) + u(2:end-1,Black_Stripes-1))/h^2;
    u(:,Black_Stripes) = Relax_Stripes_y(u(:,Black_Stripes),rhs,h,Diag_Term);
end % iterations
return

function u = Relax_Stripes_y(u,rhs,h,Diag_Term)   % 1D V cycles

% %TEST
% Diag_Term = 0;

if (size(u,1) <= 3)                     % Coarsest grid
    u(2:2:end-1,:) = (rhs(2:2:end-1,:) + (u(1:2:end-2,:) + u(3:2:end,:))/h^2) ./ ...
    (2/h^2+Diag_Term);        % One undamped Jacobi solves exactly, 
                              % because there is only one variable per stripe.
    return;
end % if

u(3:2:end-2,:) = (rhs(3:2:end-2,:) + (u(2:2:end-3,:) + u(4:2:end-1,:))/h^2) ./ ...
    (2/h^2+Diag_Term); % 1D RB Relaxation
u(2:2:end-1,:) = (rhs(2:2:end-1,:) + (u(1:2:end-2,:) + u(3:2:end,:))/h^2) ./ ...
    (2/h^2+Diag_Term); % 1D RB Relaxation


% Compute residual
L = zeros(size(rhs));
L(2:end-1,:) = (-(u(1:end-2,:) + u(3:end,:))/h^2 + u(2:end-1,:)*(2/h^2+Diag_Term));
r = rhs - L;

rhsc = 0.5*r(1:2:end,:);                % Right-hand side of coarser grid
uc = zeros(size(rhsc));                 % Zero initial guess on coarse grid
uc = Relax_Stripes_y(uc,rhsc,2*h,Diag_Term); % Recursive call
u(1:2:end,:) = u(1:2:end,:) + uc;             % Interpolate and add correction
u(2:2:end-1,:) = u(2:2:end-1,:) + 0.5*(uc(1:end-1,:) + uc(2:end,:)); % Interpolate and add correction

u(2:2:end-1,:) = (rhs(2:2:end-1,:) + (u(1:2:end-2,:) + u(3:2:end,:))/h^2) ./ ...
    (2/h^2+Diag_Term); % 1D RB Relaxation
u(3:2:end-2,:) = (rhs(3:2:end-2,:) + (u(2:2:end-3,:) + u(4:2:end-1,:))/h^2) ./ ...
    (2/h^2+Diag_Term); % 1D RB Relaxation


return

function u = Jacobi(u,f,h,nu,w,epsi)            % Jacobi relaxation with damping w

for iters = 1:nu
    u = u + w*(f - L(u,h,epsi))*h^2/(2*(1+epsi));   % Assumes f = L(u) on boundaries
end

return

function fc = Restrict(r)              % Restriction to coarser grid using 
                                       % full-weighted local averaging
fc = r(1:2:end,1:2:end);
fc(2:end-1,2:end-1) = 0.25*(fc(2:end-1,2:end-1) + ...
    0.5*(r(2:2:end-3,3:2:end-2) + r(4:2:end-1,3:2:end-2) + ...
    r(3:2:end-2,2:2:end-3) + r(3:2:end-2,4:2:end-1)) + ...
    0.25*(r(2:2:end-3,2:2:end-3) + r(2:2:end-3,4:2:end-1) + ...
    r(4:2:end-1,2:2:end-3) + r(4:2:end-1,4:2:end-1)));
return

function u = Prolong(uc)                % Bi-linear interpolation

u = zeros(length(uc)*2-1);
u(1:2:end,1:2:end) = uc;
u(2:2:end-1, 1:2:end) = 0.5*(u(1:2:end-2,1:2:end) + u(3:2:end,1:2:end));
u(1:end,2:2:end-1) = 0.5*(u(1:end,1:2:end-2) + u(1:end,3:2:end));
return

function [u,f] = Initialize_Data(n,h,epsi)

[X,Y] = meshgrid(0:h:1,0:h:1);
u = 2*sin(pi*X) + sin(2*pi*Y);
f = L(u,h,epsi);

u(2:end-1,2:end-1) = rand(n-1);  % Start with random initial guess
return